----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    01:39:21 04/03/2018 
-- Design Name: 
-- Module Name:    mz80disp - gate 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;  -- defines std_logic types
use IEEE.std_logic_unsigned.all;

entity mz80disp_gate is
	port (                             -- INPUT and OUTPUT declaration
		OSC	 	 : in  STD_LOGIC; -- input clock 16MHz
		CS	 		 : in  STD_LOGIC; -- chip select
		WR			 : in  STD_LOGIC; -- NEG cpu write
		CLK	 	 : out STD_LOGIC; -- dot clock out
		BLANK	 	 : out STD_LOGIC; -- NEG blanking time
		SYNC	    : out STD_LOGIC; -- NEG video sync
		CPU_WAIT  : out STD_LOGIC; -- NEG cpu wait
		VRAMW		 : out STD_LOGIC; -- NEG video ram write
		VRAMSEL	 : out STD_LOGIC; -- NEG video ram select(for data buffer)
		SWITCH	 : out STD_LOGIC; -- NEG video ram select
		VRAM_IN	 : out STD_LOGIC; -- video ram select(for disable output) 
		LOAD	 	 : out STD_LOGIC; -- NEG shift register(74166) data load
		ADCLK	 	 : out STD_LOGIC; -- clock for VRAM address counter 
		ROW	 	 : out STD_LOGIC_VECTOR (2 downto 0); -- cgrom row address
		LATCH		 : out STD_LOGIC :='0'; -- VRAM address latch
		PRESET	 : out STD_LOGIC; -- VRAM address preset
		RESET	 	 : out STD_LOGIC :='0' -- VRAM address reset (clear LATCH))
	);
	attribute pin_assign : string;
	attribute pin_assign of osc	   : signal is "1";
	attribute pin_assign of cs       : signal is "5";
	attribute pin_assign of wr       : signal is "6";
	attribute pin_assign of clk      : signal is "33";
	attribute pin_assign of blank    : signal is "31";
	attribute pin_assign of sync     : signal is "30";
	attribute pin_assign of cpu_wait : signal is "7";
	attribute pin_assign of vramw    : signal is "28";
	attribute pin_assign of vramsel  : signal is "29";
	attribute pin_assign of vram_in  : signal is "27";
	attribute pin_assign of switch   : signal is "19";
	attribute pin_assign of load     : signal is "32";
	attribute pin_assign of adclk    : signal is "12";
	attribute pin_assign of row      : signal is "22,21,20";
	attribute pin_assign of latch    : signal is "13";
	attribute pin_assign of preset   : signal is "14";
	attribute pin_assign of reset    : signal is "18";
end mz80disp_gate;

architecture mz80disp_gate_arch of mz80disp_gate is  -- Internal signal declaration
	signal H_BLANK      : STD_LOGIC :='0';
	signal V_BLANK      : STD_LOGIC :='0';
	signal COL          : STD_LOGIC_VECTOR (2 downto 0) :="000";
	signal H_COUNT      : STD_LOGIC_VECTOR (5 downto 0) :="000000";
	signal H_CARRY      : STD_LOGIC :='0';
	signal H_SYNC       : STD_LOGIC :='0';
	signal INT_ROW      : STD_LOGIC_VECTOR (2 downto 0) :="000";
	signal ROW_CARRY    : STD_LOGIC :='0';
	signal V_COUNT      : STD_LOGIC_VECTOR (4 downto 0) :="00000";
	signal V_SYNC       : STD_LOGIC :='0';
	signal WE		     : STD_LOGIC :='0';
	signal INT_VRAMSEL  : STD_LOGIC :='0';
	signal INT_LOAD	  : STD_LOGIC :='0';
	signal INT_CLK		  : STD_LOGIC :='0';

begin
	BLANK		<= not (H_BLANK or V_BLANK);
	CPU_WAIT	<= CS nand (not WE);
	INT_VRAMSEL <= CS nand V_BLANK;
	VRAMSEL	<= INT_VRAMSEL;
	VRAM_IN	<= not INT_VRAMSEL;
	SWITCH	<= INT_VRAMSEL;
	VRAMW		<= INT_VRAMSEL or WR;
	SYNC		<= not(H_SYNC xor V_SYNC);
	ROW		<= INT_ROW;
   LOAD		<= not INT_LOAD;
   ADCLK		<= not INT_LOAD;
	PRESET	<= H_CARRY;
	CLK		<= INT_CLK;

	T_FF: process (OSC) -- devide to 8MHz
	begin
		if (OSC'event and OSC='1') then
			if (INT_CLK = '0') then
				INT_CLK <= '1';
			else
				INT_CLK <= '0';
			end if;
		end if;
	end process;

	COL_COUNTER: process (INT_CLK) -- octal counter
	begin
		if (INT_CLK'event and INT_CLK='0') then
			COL <= COL + 1;
			if (COL = "111") then
				INT_LOAD <= '1';  --carry / shift register data load
			else
				INT_LOAD <= '0';
			end if;
		end if;
	end process;

	H_COUNTER: process (INT_LOAD) -- 6bit 64 binary counter
	begin
		if (INT_LOAD'event and INT_LOAD='1') then
			H_COUNT <= H_COUNT + 1;
			if (H_COUNT = "000000") then	  --64
				H_CARRY <= '1';				--preset
				LATCH <= '0';
			else
				H_CARRY <= '0';
			end if;
			if (H_COUNT = "000001") then	  --1
				H_BLANK <= '0';
			end if;
			if (H_COUNT = "101001") then    --41
				H_BLANK <= '1';
				if(INT_ROW = "111") then
						LATCH <= '1';			--latch
				end if;
			end if;
			if (H_COUNT = "110010") then    --50
				H_SYNC <= '1';
			end if;
			if (H_COUNT = "110110") then    --54
				H_SYNC <= '0';
			end if;
		end if;
	end process;

	ROW_COUNT: process (H_CARRY) -- octal counter
	begin
		if (H_CARRY'event and H_CARRY='1') then
			INT_ROW <= INT_ROW + 1;
			if (INT_ROW = "111") then
				ROW_CARRY <= '1';
			else
				ROW_CARRY <= '0';
			end if;
		end if;
	end process;

	V_COUNTER: process (ROW_CARRY) -- 5bit 32 binary counter
	begin
		if (ROW_CARRY'event and ROW_CARRY='1') then
			V_COUNT <= V_COUNT + 1;
			if (V_COUNT = "00000") then     --32
				V_BLANK <= '0';
				RESET	  <= '0';
			end if;
			if (V_COUNT = "11001") then    --25
				V_BLANK <= '1';
				WE		  <= '1';
			end if;
			if (V_COUNT = "11011") then    --27
				V_SYNC <= '1';
			end if;
			if (V_COUNT = "11100") then    --28
				V_SYNC <= '0';
			end if;
			if (V_COUNT = "11111") then    --31
				WE		 <= '0';
				RESET	  <= '1';				--reset (clear LATCH)
			end if;
		end if;
	end process;

end mz80disp_gate_arch;