/********************************************************/
/* AKI-80 VRAM BOARD LIBRARY                            */
/*                                 source for HI-TECH C */
/* 2020.07.14                             (c)neko Java  */
/********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include "vrmlib.h"
#include "vrmfont.h"

/* global for Assembler **********************************/
char Vram_param = 0;		/* VRAM control parameter */
unsigned int Vram_ad = 0;	/* for source address */
unsigned int Vram_ad2 = 0;	/* for destination address */
unsigned int Vram_rep = 0;
char Vram_data = 0;
char Vram_pcg[8];
unsigned int Vram_cur = 0;
char Vram_buff[VRM_COUNT];


static void vram_set_text_mode(void)
{
	Vram_param |= VRM_TX;
}

static void vram_set_graph_mode(void)
{
	Vram_param &= VRM_GR;
}

static void vram_set_page(char page)	/* 0-3 */
{
	page %= 4;
	page <<= 4;
	Vram_param &= VRM_PG;
	Vram_param |= page;
}

static void vram_set_row(char row)	/* 0-7 */
{
	row %= 8;
	Vram_param &= VRM_ROW;
	Vram_param |= row;
}

static void vram_io_write(void)
{
#asm
	PUSH	AF
	LD		A,(_Vram_param)
	SET		7,A
	OUT		(VRM_IOPT),A
	POP		AF
#endasm
}

static void vram_write(void)
{
#asm
	PUSH 	AF
	PUSH	HL
	LD		A,(_Vram_param)
	OUT		(VRM_IOPT),A
	EX		AF,AF'
	LD		A,(_Vram_data)
	LD		HL,(_Vram_ad)
	LD		(HL),A
	EX		AF,AF'
	SET		7,A
	OUT		(VRM_IOPT),A
	POP		HL
	POP		AF
#endasm
}

static void vram_bit_set(void)
{
#asm
	PUSH 	AF
	PUSH 	BC
	PUSH	HL
	LD		A,(_Vram_param)
	OUT		(VRM_IOPT),A
	EX		AF,AF'
	LD		A,(_Vram_data)
	LD		HL,(_Vram_ad)
	OR		(HL)
	LD		(HL),A
	EX		AF,AF'
	SET		7,A
	OUT		(VRM_IOPT),A
	POP		HL
	POP		BC
	POP		AF
#endasm
}

static void vram_bit_reset(void)
{
#asm
	PUSH 	AF
	PUSH 	BC
	PUSH	HL
	LD		A,(_Vram_param)
	OUT		(VRM_IOPT),A
	EX		AF,AF'
	LD		A,(_Vram_data)
	LD		HL,(_Vram_ad)
	AND		(HL)
	LD		(HL),A
	EX		AF,AF'
	SET		7,A
	OUT		(VRM_IOPT),A
	POP		HL
	POP		BC
	POP		AF
#endasm
}

static void vram_pcg_write(void)
{
#asm
	PUSH	AF
	PUSH	BC
	PUSH	DE
	PUSH	HL
	LD		B,8H
	LD		HL,_Vram_pcg
	LD		DE,(_Vram_ad)
	LD		A,(_Vram_param)
	PUSH	AF
PCGLP:
	OUT		(VRM_IOPT),A
	EX		AF,AF'
	LD		A,(HL)
	LD		(DE),A
	INC		HL
	EX		AF,AF'
	INC		A
	DJNZ	PCGLP
	POP		AF
	SET		7,A
	OUT		(VRM_IOPT),A
	POP		HL
	POP		DE
	POP		BC
	POP		AF
#endasm
}

static void vram_pcg_write_8kb(void)
{
#asm
	PUSH	AF
	PUSH	BC
	PUSH	DE
	PUSH	HL
	LD		B,8H
	LD		HL,_Vram_pcg
	LD		DE,(_Vram_ad)
	LD		A,(_Vram_param)
	PUSH	AF
PCGLP8K:
	OUT		(VRM_IOPT),A
	EX		AF,AF'
	LD		A,(HL)
	LD		(DE),A
	INC		HL
	LD		A,D
	ADD		A,4
	LD		D,A
	EX		AF,AF'
	INC		A
	DJNZ	PCGLP8K
	POP		AF
	SET		7,A
	OUT		(VRM_IOPT),A
	POP		HL
	POP		DE
	POP		BC
	POP		AF
#endasm
}

static void vram_pcg_right(void) /* 1dot shift right */
{
#asm
	PUSH	AF
	PUSH	BC
	PUSH	DE
	PUSH	HL
	LD		B,8H
	LD		HL,(_Vram_ad)
	EXX
	LD		HL,(_Vram_ad)
	EXX
	LD		DE,(_Vram_rep)
	LD		A,(_Vram_param)
	PUSH	AF
PCGSR:
	OUT		(VRM_IOPT),A
	LD		C,E
	EX		AF,AF'
	XOR		A
PCGSR1:
	LD		A,(HL)
	RRA
	LD		(HL),A
	INC		HL
	DEC		C
	JR		NZ,PCGSR1
	EXX
	LD		A,H
	ADD		A,4
	LD		H,A
	EXX
	LD		H,A
	EXX
	LD		A,L
	EXX
	LD		L,A
	EX		AF,AF'
	INC		A
	DJNZ	PCGSR
	POP		AF
	SET		7,A
	OUT		(VRM_IOPT),A
	POP		HL
	POP		DE
	POP		BC
	POP		AF
#endasm
}

static void vram_pcg_left(void) /* 1dot shift left */
{
#asm
	PUSH	AF
	PUSH	BC
	PUSH	DE
	PUSH	HL
	LD		B,8H
	LD		HL,(_Vram_ad)
	EXX
	LD		HL,(_Vram_ad)
	EXX
	LD		DE,(_Vram_rep)
	LD		A,(_Vram_param)
	PUSH	AF
PCGSL:
	OUT		(VRM_IOPT),A
	LD		C,E
	EX		AF,AF'
	XOR		A
PCGSL1:
	LD		A,(HL)
	RLA
	LD		(HL),A
	DEC		HL
	DEC		C
	JR		NZ,PCGSL1
	EXX
	LD		A,H
	ADD		A,4
	LD		H,A
	EXX
	LD		H,A
	EXX
	LD		A,L
	EXX
	LD		L,A
	EX		AF,AF'
	INC		A
	DJNZ	PCGSL
	POP		AF
	SET		7,A
	OUT		(VRM_IOPT),A
	POP		HL
	POP		DE
	POP		BC
	POP		AF
#endasm
}

static void vram_zero(void)
{
#asm
	PUSH	AF
	PUSH	BC
	PUSH	DE
	PUSH	HL
	LD		A,(_Vram_param)
	OUT		(VRM_IOPT),A
	LD		HL,(_Vram_ad)
	LD		(HL),0
	LD		DE,(_Vram_ad2)
	LD		BC,(_Vram_rep)
	LDIR
	SET		7,A
	OUT		(VRM_IOPT),A
	POP		HL
	POP		DE
	POP 	BC
	POP		AF
#endasm
}

static void vram_move_data(void)
{
#asm
	PUSH	AF
	PUSH	BC
	PUSH	DE
	PUSH	HL
	LD		A,(_Vram_param)
	OUT		(VRM_IOPT),A
	LD		HL,(_Vram_ad)
	LD		DE,(_Vram_ad2)
	LD		BC,(_Vram_rep)
	LDIR
	SET		7,A
	OUT		(VRM_IOPT),A
	POP		HL
	POP		DE
	POP 	BC
	POP		AF
#endasm
}

static void vram_read_data(void)
{
#asm
	PUSH 	AF
	PUSH	HL
	LD		HL,(_Vram_ad)
	LD		A,(HL)
	LD		(_Vram_data),A
	POP		HL
	POP		AF
#endasm
}

void vram_text(char page, char row)
{
	vram_set_text_mode();
	vram_set_page(page);
	vram_set_row(row);
	
	vram_io_write();
}

void vram_graph(char page)
{
	vram_set_graph_mode();
	vram_set_page(page);
	vram_set_row(0);
	
	vram_io_write();
}

static void vram_clear_text(void)
{
	Vram_ad = VRM_BASE;
	Vram_ad2 = VRM_BASE+1;
	Vram_rep = VRM_COUNT;
	
	vram_zero();
	Vram_cur = 0;
}

static void vram_clear_graph(void)	/* clear 8kb area for graph bit operation */
{
	int i;
	Vram_ad = VRM_BASE;
	Vram_ad2 = VRM_BASE+1;
	Vram_rep = VRM_COUNT;
	
	for (i=0; i<8; i++){
		vram_zero();
		Vram_param++;
		Vram_ad += VRM_1KB;
		Vram_ad2 += VRM_1KB;
	}
	vram_set_graph_mode();
}

void vram_clear(void)
{
	if(Vram_param & VRM_TX) vram_clear_text(); else vram_clear_graph();
}

void vram_clear_1kb(void)	/* clear graph but use only 1kb area */
{
	int i;
	Vram_ad = VRM_BASE;
	Vram_ad2 = VRM_BASE+1;
	Vram_rep = VRM_COUNT;
	
	for (i=0; i<8; i++){
		vram_zero();
		Vram_param++;
	}
	vram_set_graph_mode();
}

void vram_text_scroll(void)
{
	Vram_ad = VRM_BASE+VRM_TX_COL;
	Vram_ad2 = VRM_BASE;
	Vram_rep = VRM_COUNT-VRM_TX_COL;
	
	vram_move_data();
	
	Vram_ad = VRM_BASE+VRM_TX_COL*(VRM_TX_ROW-1);
	Vram_ad2 = Vram_ad+1;
	Vram_rep = VRM_TX_COL-1;
	
	vram_zero();
	Vram_cur = VRM_COUNT - VRM_TX_COL;
}

void vram_pix_set(int x, int y)
{
	if(x<0 || x>=VRM_GR_X || y<0 || y>=VRM_GR_Y) return;
	Vram_param += y%8;
	Vram_data = 0x80 >> x%8;
	Vram_ad = VRM_BASE + VRM_1KB*(y%8) + x/8 + (y/8)*VRM_TX_COL;
	vram_bit_set();
	vram_set_row(0);
}

void vram_pix_reset(int x, int y)
{
	if(x<0 || x>=VRM_GR_X || y<0 || y>=VRM_GR_Y) return;
	Vram_param += y%8;
	Vram_data = ~(0x80 >> x%8);
	Vram_ad = VRM_BASE + VRM_1KB*(y%8) + x/8 + (y/8)*VRM_TX_COL;
	vram_bit_reset();
	vram_set_row(0);
}

void vram_pix_bw(int x, int y, char bw)		/* 1:white, 0:black */
{
	if(bw) vram_pix_set(x, y); else vram_pix_reset(x, y);
}

void vram_line_bw(int x1, int y1, int x2, int y2, char bw)	/* 1:white, 0:black */
{
	int dx, dy, s, step;
	
	if (x1==x2 && y1==y2) return;

	dx = abs(x2 - x1);  dy = abs(y2 - y1);
	if (dx > dy) {
		if (x1 > x2) {
			step = (y1 > y2) ? 1 : -1;
			s = x1;  x1 = x2;  x2 = s;  y1 = y2;
		} else step = (y1 < y2) ? 1: -1;
		vram_pix_bw(x1, y1, bw);
		s = dx >> 1;
		while (++x1 <= x2) {
			if ((s -= dy) < 0) {
				s += dx;  y1 += step;
			};
			vram_pix_bw(x1, y1, bw);
		}
	} else {
		if (y1 > y2) {
			step = (x1 > x2) ? 1 : -1;
			s = y1;  y1 = y2;  y2 = s;  x1 = x2;
		} else step = (x1 < x2) ? 1 : -1;
		vram_pix_bw(x1, y1, bw);
		s = dy >> 1;
		while (++y1 <= y2) {
			if ((s -= dx) < 0) {
				s += dy;  x1 += step;
			}
			vram_pix_bw(x1, y1, bw);
		}
	}
}

void vram_line(int x1, int y1, int x2, int y2)
{
	vram_line_bw(x1, y1, x2, y2, 1);
}

void vram_bline(int x1, int y1, int x2, int y2)
{
	vram_line_bw(x1, y1, x2, y2, 0);
}

void vram_rect(int x1, int y1, int x2, int y2)
{
	vram_line(x1, y1, x2, y1);
	vram_line(x2, y1, x2, y2);
	vram_line(x2, y2, x1, y2);
	vram_line(x1, y2, x1, y1);
}

void vram_circle(int xc, int yc, int r, char bw)
{
	int x, y;
	x = r;  y = 0;
	while (x >= y) {
		vram_pix_bw(xc + x, yc + y, bw);
		vram_pix_bw(xc + x, yc - y, bw);
		vram_pix_bw(xc - x, yc + y, bw);
		vram_pix_bw(xc - x, yc - y, bw);
		vram_pix_bw(xc + y, yc + x, bw);
		vram_pix_bw(xc + y, yc - x, bw);
		vram_pix_bw(xc - y, yc + x, bw);
		vram_pix_bw(xc - y, yc - x, bw);
		if ((r -= (y++ << 1) - 1) < 0) r += (x-- - 1) << 1;
	}
}

void vram_cur_set(unsigned int x, unsigned int y)
{
	x %=VRM_TX_COL; y %=VRM_TX_ROW;
	
	Vram_cur = x+y*VRM_TX_COL;
}

void vram_cur_inc(void)
{
	Vram_cur++;
}

void vram_cur_cr(void)
{
	if(Vram_cur >= (VRM_COUNT-VRM_TX_COL)) vram_text_scroll();
	else Vram_cur += VRM_TX_COL - Vram_cur % VRM_TX_COL;
}

void vram_putdc(char c)	/* put direct display code */
{
	if(Vram_cur == VRM_COUNT) vram_text_scroll();
	Vram_ad = VRM_BASE+Vram_cur;
	Vram_data = c;
	vram_write();
	vram_cur_inc();
}

char vram_chg_code(char c)	/* from ascii code to display code */
{
	if(c>=0    && c<=0x10) return 0;
	if(c>=0x11 && c<=0x16) return c + 0xB0;
	if(c>=0x17 && c<=0x20) return 0;
	if(c>=0x21 && c<=0x29) return c + 0x40;
	if(c==0x2A) return 0x6B;
	if(c==0x2B) return 0x6A;
	if(c==0x2C) return 0x2F;
	if(c==0x2D) return 0x2A;
	if(c==0x2E) return c;
	if(c==0x2F) return 0x2D;
	if(c>=0x30 && c<=0x39) return c - 0x10;
	if(c==0x3A) return 0x4F;
	if(c==0x3B) return 0x2C;
	if(c==0x3C) return 0x51;
	if(c==0x3D) return 0x2B;
	if(c==0x3E) return 0x57;
	if(c==0x3F) return 0x49;
	if(c==0x40) return 0x55;
	if(c>=0x41 && c<=0x5A) return c - 0x40;
	if(c==0x5B) return 0x52;
	if(c==0x5C) return 0x59;
	if(c==0x5D) return 0x54;
	if(c>=0x61 && c<=0x7A) return c - 0x60;
	return 0;
}

void vram_putch(char c)
{
	if(c==0x0D){ vram_cur_cr(); return;}
	if(Vram_cur == VRM_COUNT) vram_text_scroll();
	
	Vram_ad = VRM_BASE+Vram_cur;
	Vram_data = vram_chg_code(c);
	vram_write();
	vram_cur_inc();
}

void vram_print(char *str)
{
	int i;
	for(i=0; i<strlen(str); i++){
		vram_putch(str[i]);
	}
}

void vram_gput88(char *ptn)	/* put 8*8 bit patern */
{
	if(Vram_cur >= VRM_COUNT) Vram_cur=0;
	Vram_ad = VRM_BASE+Vram_cur;
	memcpy(Vram_pcg, ptn, 8);
	vram_pcg_write();
	vram_cur_inc();
}

void vram_gput88_8kb(char *ptn)	/* put 8*8 bit patern. possible bit operation */
{
	if(Vram_cur >= VRM_COUNT) Vram_cur=0;
	Vram_ad = VRM_BASE+Vram_cur;
	memcpy(Vram_pcg, ptn, 8);
	vram_pcg_write_8kb();
	vram_cur_inc();
}

void vram_g88_shift(int blocks, int bits) /* 8*8 bit patter shift. bits +:right/ -:left */ 
{
	int i;
	
	if (bits==0) return;
	if(Vram_cur >= VRM_COUNT) Vram_cur=0;
	Vram_ad = VRM_BASE+Vram_cur;
	Vram_rep = blocks;
	
	if(bits>0){
		for(i=0; i<bits; i++) vram_pcg_right();
	}else{
		bits *= -1;
		for(i=0; i<bits; i++) vram_pcg_left();
	}
}

void vram_gputch(char c)
{
	if(Vram_cur >= VRM_COUNT) Vram_cur=0;
	if(c<0x20 || c>0x7F) c = 0x20;
	Vram_ad = VRM_BASE+Vram_cur;
	memcpy(Vram_pcg, Vram_font+(c-0x20)*8, 8);
	vram_pcg_write();
	vram_cur_inc();
}

void vram_gprint(char *str)
{
	int i;
	for(i=0; i<strlen(str); i++){
		vram_gputch(str[i]);
	}
}

char vram_read(int row)	 /* 1byte vram read data. cursor base+row */
{
	row %=8;
	if(Vram_cur >= VRM_COUNT) Vram_cur=0;
	Vram_ad = VRM_BASE+VRM_1KB*row+Vram_cur;
	vram_read_data();
	return Vram_data;
}

int vram_disp_img(char *fname)	/* Windows bitmap 320*200 1bit format */
{
	int i, j;
	FILE* fp = fopen(fname, "rb");
	
	Vram_ad = (int)Vram_buff;
	Vram_ad2 = VRM_BASE;
	Vram_rep = VRM_COUNT;
	
	if(fp == NULL) return 1;
	for(j=0; j<8; j++){
		if(fseek(fp, 62+VRM_COUNT*8-VRM_TX_COL*j, 0) != 0) {fclose(fp); return 2;}
		for(i=0; i<VRM_TX_ROW; i++){
			if(fseek(fp, -VRM_TX_COL, 1) != 0) {fclose(fp); return 3;}
			if(fread(Vram_buff+VRM_TX_COL*i, sizeof(char), VRM_TX_COL, fp) < 1){fclose(fp); return 4;}
			if(i==(VRM_TX_ROW-1)) break;
			if(fseek(fp, -VRM_TX_COL*8, 1) != 0) {fclose(fp); return 5;}
		}
		vram_set_row(j);
		vram_move_data();
	}
	if(fclose(fp) == EOF) return 6;
	return 0;
}
